/**
A combination (or state) associates a {@link Value} with each
{@link Variable}.
@author Michael Breen
*/
public class Combination
{
   /**
   The number of possible combinations based on the
   variables.
   */
   private static long count = 0;

   /**
   Essentially the inverse of {@link #getNumber()}.
   All distinct combinations are associated with a unique
   number, that is, two combinations have the same number
   only if they are equal.
   @param number a number in the range
   0 .. ({@link #numCombinations()} - 1)
   @return a combination whose {@link #getNumber()}
   method returns the argument
   */
   public static Combination getCombination(long number)
   {
      if (number < 0 || number >= numCombinations())
         throw new IndexOutOfBoundsException();
      Combination comb = new Combination();
      for (int i = Variable.numVars() - 1; i >= 0; --i)
      {
         Variable var = Variable.getVariable(i);
         int numVals = var.numVals();
         comb.setValue(var.getValue((int) number % numVals));
         number /= numVals;
      }
      return comb;
   }

   /**
   Get the number of distinct combinations that can be
   formed, ignoring constraints.
   This is simply the product of the number of values of
   each variable.
   @return the number of possible combinations
   */
   public static long numCombinations()
   {
      if (Combination.count == 0)
      {
         Combination.count = 1;
         for (int i = Variable.numVars() - 1; i >= 0; --i)
         {
            Combination.count *=
               Variable.getVariable(i).numVals();
         }
      }
      return Combination.count;
   }

   /**
   For convenience, the combination is implemented
   as a mask: where the mask allows a variable to
   have any of several values, the variable's value
   is taken to be the first such value in definition
   order.
   */
   private CombinationMask mask;

   /**
   Each variable in the new combination is assigned
   the first value defined for that variable.
   */
   public Combination() { this.mask = new CombinationMask(); }

   /**
   The value of each variable is the first defined
   value which it is allowed in the argument.
   @param allowed a mask allowing each variable to
   have one of a set of one or more values
   */
   Combination(CombinationMask allowed)
   {
      this.mask = new CombinationMask(allowed);
   }

   /** The new object is a copy (clone) of the argument. */
   public Combination(Combination org)
   {
      this();
      setEqualTo(org);
   }

   /**
   Find the value of a given variable in this object.
   @param var a variable in the model
   @return the value of the specified variable
   @see #setValue
   */
   public Value getValue(Variable var)
   {
      return (Value)
         this.mask.getValues(var).iterator().next();
   }

   /**
   Set the value of a variable in this object.
   The variable does not need to be specified as it
   can be determined from the value.
   @param value the new value of a variable
   @see #getValue
   */
   public void setValue(Value value)
   {
      this.mask.setValues(value);
   }

   /**
   Make this object equal to the argument.
   @param comb a combination to copy from
   */
   void setEqualTo(Combination comb)
   {
      this.mask = new CombinationMask(comb);
   }

   /**
   Two combinations are equal if the values of the
   variables are the same in both.
   @param obj the object to be compared with this object
   @return true if the values of the variables in the
   argument are the same as in this object, otherwise false
   */
   public boolean equals(Object obj)
   {
      if (this == obj)
         return true;
      if (!(obj instanceof Combination))
         return false;
      return getNumber() == ((Combination) obj).getNumber();
   }

   /**
   Every distinct combination produces a unique hash code
   (to the extent allowed by the resolution of an integer).
   */
   public int hashCode() { return (int) getNumber(); }

   /**
   Get the number uniquely associated with the combination
   of variable values stored in this object.
   @return a number in the range 0 ..
   ({@link #numCombinations()} - 1) that uniquely identifies
   the combination of variable values stored in this object
   */
   public long getNumber()
   {
      long result = 0;
      long factor = 1;
      for (int i = Variable.numVars() - 1; i >= 0; --i)
      {
         Variable var = Variable.getVariable(i);
         result += factor * getValue(var).getIndex();
         factor *= var.numVals();
      }
      return result;
   }

   /**
   Produces a string in the form, e.g.,
   "{Door = open, Motor = off, ...}"
   @return a string representation of this object
   */
   public String toString()
   {
      StringBuffer buf = new StringBuffer("{");
      for (int i = 0; i < Variable.numVars(); ++i)
      {
         if (i > 0)
            buf.append(", ");
         buf.append(getValue(Variable.getVariable(i)));
      }
      buf.append("}");
      return buf.toString();
   }
<#list usable_variables as var>

   /**
   Get the value of variable '${var.@name}'.
   @return the value of '${var.@name}' in this combination
   @see #setValue
   */
   public ${variable_id(var.@name)} get${variable_id(var.@name)}()
   {
      return (${variable_id(var.@name)}) <#--
-->getValue(Variable.${variable_id(var.@name)});
   }
</#list>
}
