import java.util.HashMap;
import java.util.Iterator;

/**
A combination mask associates a value or a set of values
(possibly "don't care") with each variable.
A combination is allowed by a mask if the value of
each variable in the combination is one of the values
in the mask.

@author Michael Breen
*/
class CombinationMask
{
   /**
   Maps variables to sets of their values;
   unmapped variables are implicitly "don't care".
   */
   HashMap valueSets = new HashMap();

   /**
   Creates a mask allowing all combinations, that is,
   all variables are "don't care".
   */
   CombinationMask() { }

   /** The new mask is a copy (clone) of the argument. */
   CombinationMask(CombinationMask org)
   {
      this.valueSets.putAll(org.valueSets);
   }

   /** The new mask allows only the argument. */
   CombinationMask(Combination comb)
   {
      for (int i = 0; i < Variable.numVars(); ++i)
      {
         Variable var = Variable.getVariable(i);
         this.valueSets.put(var, comb.getValue(var));
      }
   }

   /**
   @param comb the combination being tested
   @return true if the argument is allowed by this mask,
   otherwise false
   @throws NullPointerException if the argument is null
   */
   boolean allows(Combination comb)
   {
      Iterator iter = this.valueSets.keySet().iterator();
      while (iter.hasNext())
      {
         ValueSet allowed = (ValueSet)
            this.valueSets.get(iter.next());
         if (!comb.getValue(allowed.getVariable()).in(allowed))
         {
            return false;
         }
      }
      return true;
   }

   /**
   @param variable any variable
   @return the values the variable is allowed by this mask
   */
   ValueSet getValues(Variable variable)
   {
      ValueSet vals = (ValueSet) this.valueSets.get(variable);
      if (vals == null)
         vals = variable.dontCare();
      return vals;
   }

   /**
   @param values the set of values the corresponding variable
   is to be allowed by this mask
   @return this object (to facilitate invocation chaining)
   @throws IllegalArgumentException if the argument is null
   */
   CombinationMask setValues(ValueSet values)
   {
      if (values == null)
         throw new IllegalArgumentException();
      this.valueSets.put(values.getVariable(), values);
      return this;
   }
}

