/**
A rule from the model (table and formula parts).

@author Michael Breen
*/
public class Rule
{
   /** Stores the logic associated with a rule formula. */
   private static class Formula
   {
      /**
      @param comb the current values of the variables
      @param next any restriction on the new values of the
      variables after the rule is applied (the rule will
      not apply unless it can satisfy this)
      @return any combination of new variable values that
      are consistent with the formula and the arguments;
      null if the formula is not consistent with the arguments
      */
      Combination apply(Combination comb, CombinationMask next)
      {
         return new Combination(next);
      }
   }

   /** The event to which this rule responds. */
   private Event event;

   /** The name of this rule as defined in the model. */
   private String id;

   /** The source row of the table part of the rule. */
   private CombinationMask source = new CombinationMask();

   /** The target row of the table part of the rule. */
   private TargetRow target = new TargetRow();

   /** The logic of the rule's formula; null if it has none. */
   private Formula formula = null;

   /**
   The target and source rows of the table part of the
   new rule are initially filled with "don't cares"
   and "no changes" respectively.
   @param event the event to which this rule defines a
   response
   @param name the identifier for this rule in the model
   */
   private Rule(Event event, String name)
   {
      this.event = event;
      this.id = name;
      Rules.addRule(this);
   }

   /**
   Get the name of this rule in the model.
   Note that two rules associated with different
   events may have the same name.
   @return this rule's identifier in the model
   */
   public String getName() { return this.id; }

   /**
   Produces a string in the form
   "[<em>event_name</em>]&nbsp;<em>rule_name</em>".
   @return this identifier of this rule in the model,
   preceded by the event it is associated with
   */
   public String toString()
   {
      return "[" + this.event + "] " + this.id;
   }

   /**
   Get the event to which this rule defines a response.
   @return the event to which this rule responds
   */
   public Event getEvent() { return this.event; }

   /**
   Check whether this rule allows a transition between
   two specific states.
   @return true if this rule allows a transition from the
   specified current state to the specified next state,
   otherwise false
   */
   public boolean applies(Combination current, Combination next)
   {
      if (!this.source.allows(current))
         return false;
      CombinationMask nextMask = this.target.transform(current);
      if (!nextMask.allows(next))
         return false;
      if (this.formula != null)
      {
         return this.formula.apply(current,
            new CombinationMask(next)) != null;
      }
      return true;
   }

   /**
   @param values values to put in the source row of the rule
   @return this object (to facilitate invocation chaining)
   */
   Rule setCurrent(ValueSet values)
   {
      this.source.setValues(values);
      return this;
   }

   /**
   @param values values to put in the target row of the rule
   @return this object (to facilitate invocation chaining)
   */
   Rule setNext(ValueSet values)
   {
      this.target.setValues(values);
      return this;
   }

   /**
   @param formula the formula logic for this rule
   @return this object (to facilitate invocation chaining)
   */
   Rule setFormula(Formula formula)
   {
      this.formula = formula;
      return this;
   }

   /**
   Given a current state to which this rule applies,
   find out the new state to which the rule causes
   a transition. If there is more than one possible
   next state then one is chosen arbitrarily.
   This method can also be used to discover whether
   the rule applies in the current state.
   @param current the current values of the variables
   @return a combination of new variable values to
   which the rule maps the argument;
   null if the rule does not apply to the argument
   */
   Combination apply(Combination current)
   {
      if (!this.source.allows(current))
         return null;
      CombinationMask nextMask = this.target.transform(current);
      if (this.formula != null)
         return this.formula.apply(current, nextMask);
      else
         return new Combination(nextMask);
   }
<#list doc.model.events.event as event>

   /* Rules for Event '${event.@name}': */
   <#list event.rules.rule as rule>

   public static final Rule ${rule_id(rule)} = new Rule(
      Event.${event_id(event)}, "${rule.@name}")<#--
-->
      <@source_init rule/>
      <@target_init rule/>
      <@formula_init rule/>
;
   </#list>
</#list>
}
<#--------------------  Local Macros  -------------------->

<#-- output code to initialise the source row of a rule -->
<#macro source_init rule_node>
   <#list rule_node["./values_array/value_set[@type = 'source']"]
      as source_vals
   >
      <#--
      "Don't care" is the default in the code for the
      source row so no need to output anything for that.
      -->
      <#if source_vals.@content != "dont_care">

      .setCurrent(<@values_to_code source_vals/>)<#--
-->
      </#if>
   </#list>
</#macro>

<#-- output code to initialise the target row of a rule -->
<#macro target_init rule_node>
   <#list usable_variables as var>
      <#attempt>
         <#--
         this works only if values are specified for 'var',
         otherwise the recover block is executed...
         -->
         <#local vals = rule_node[
"./values_array/value_set[@type='target'][@variable='${var.@name}']"]
          [0]
         />
         <#--
         "no change" is the default in the code and so
         requires no output
         -->
         <#if vals.@content != "no_change">

      .setNext(<@values_to_code vals/>)<#--
-->
         </#if>
      <#recover>
         <#--
         if no values are listed for a variable in the
         model file the default semantics is "don't care"
         -->

      .setNext(${dont_care_id(var.@name)})<#--
-->
      </#recover>
   </#list>
</#macro>

<#-- output code to initialise the formula of a rule -->
<#macro formula_init rule_node>
   <#if rule_node.formula?size != 0>

      .setFormula(new Formula()
      {
         <@comment text=rule_node.formula.text
          indent="         "
         />
      <#if rule_node.formula.parse_node?size != 0>
         Combination apply(Combination comb,
            CombinationMask next)
         {
         <#-- if there are any assignments in the formula -->
         <#if rule_node.formula[
          ".//value_set[@type='target']"]?size != 0
         >
            <#-- then we'll be using these variables -->
            ValueSet mask, restricted, excluded;
         </#if>
         <@parse_node_to_code
          node = rule_node.formula.parse_node
          fail_cmd = "return null;"
         />
            return new Combination(next);
         }
      </#if>
      })<#--
-->
   </#if>
</#macro>
